// efg, Stowers Institute, 15 Aug 2003. Modified 27 Apr 2004.
//
// Modified 15 February 2006 to support Float $DataType used
// with BD (Becton, Dickinson) FACSDiva Software for
// See http://facs.stanford.edu/sff/doc/bdfacsdivasoftwareuserguide.pdf
//
// Released under GNU General Public License
//
// References:
//
// Data File Standard for Flow Cyometry, Version FCS 3.0
//      http://greenhorn.stc.cmu.edu/FCSAPI/FCS3.html
//
// Seamer, L.C, et al, Proposed New Data File Standard for Flow Cytometry,
//      Version FCS 3.0, Cytometry, 28:118-122, 1997.
//      http://greenhorn.stc.cmu.edu/publications/64-seamer1997.pdf
//
// Leif, Robert C and Suzanne B. Leif, The Evolution of Flow Cytometry
//      Standard, FCS3.0, into a DICOM Compatible Format, Optical
//      Diagnostics of Biological Fluids and Advanced Techniques in
//      Analytical Cytogy, SPIE Procs Series, Vol 2982, pp. 1-13, 1997.
//      http://newportinstruments.com/cytometryml/fcs_to_dicom1.pdf
//  *** petite modif Jean-françois Madre le 19/02/13 pour écrire un message si le format de fichier n'est pas lu ***
//  *** modifs pour lire quelques paramètres supplémentaires et pour la lecture des fichiers avec des données de type Entier (I)
//      lorsque le codage n'est pas sur 2 bits
// remplacement des Assert par des MessageDLg
UNIT FCSLibrary;

{$MODE Delphi}

INTERFACE

USES
  Classes,     // TFileStream
  Grids,       // TStringGrid
  SysUtils;    // TFileName
    //Donnees,
TYPE
  TLigneDonnees = Array of Real;
  TFCSFile =
  CLASS(TObject)
    PRIVATE
      FFileName  :  TFileName;
      FFileStream:  TFileStream;

      FHeaderTextFirst    :  INTEGER;   // Bytes 10-17  (0-based)
      FHeaderTextLast     :  INTEGER;   // Bytes 18-25
      FHeaderDataFirst    :  INTEGER;   // Bytes 26-33
      FHeaderDataLast     :  INTEGER;   // Bytes 34-41
      FHeaderAnalysisFirst:  INTEGER;   // Bytes 42-49
      FHeaderAnalysisLast :  INTEGER;   // Bytes 50-57

      FDataElementSize    :  INTEGER;   // 1 to 4 for 'I' or 4 for 'F' DATATYPE

      FCS_PROJ_ProjectName         :  STRING;
      FCS_DATATYPE                 :  STRING;
      FCS_CELLS_TypeOfCellsMeasured:  STRING;
      FCS_MODE                     :  STRING;
      FCS_TOT_EventCount           :  LongInt;
      FCS_PAR_ParameterCount       :  LongInt;

      FCS_BYTEORD                  :  STRING;   // Assumed:  1,2,3,4 (PC)
      FBigEndian                   :  BOOLEAN;  // Assumed FALSE

      FVersion                     :  STRING;
      FCS_Cytometer                :  STRING;
      FCS_LaDate                   :  STRING;

      FName                        :  ARRAY OF STRING;
      FNameString                  :  ARRAY OF STRING;
      FRange                       :  ARRAY OF INTEGER;
      FMask                        :  ARRAY OF INTEGER;
      FBits                        :  ARRAY OF INTEGER;
      FParamE1                     :  Array of integer; // ajout JFM
      FParamE2                     :  Array of real; // ajout JFM
      FSelect                      :  ARRAY OF BOOLEAN;  // output selections

      FUNCTION GetName  (index:  INTEGER):  STRING;
      FUNCTION GetNameString(index:  INTEGER):  STRING;
      FUNCTION GetRange (index:  INTEGER):  INTEGER;
      FUNCTION GetBits  (index:  INTEGER):  INTEGER;

      FUNCTION  GetSelect (index:  INTEGER):  BOOLEAN;
      PROCEDURE SetSelect (index:  INTEGER; value:  BOOLEAN);
      FUNCTION GETParamE1 (index:  INTEGER) : integer;
      FUNCTION GetParamE2 (index:  Integer) : real;
    PUBLIC
      //CONSTRUCTOR Create(CONST Filename:  TFileName);
      CONSTRUCTOR Create(CONST Filename:  string);
      DESTRUCTOR  Destroy; OVERRIDE;

      FUNCTION  GetTextData:  STRING;
      FUNCTION  GetNameRow(CONST NameFormat:  INTEGER):  STRING;
      FUNCTION  GetDataRow(CONST Row:  INTEGER;
                           CONST ApplyMasking, FixEndianess:  BOOLEAN):  STRING;
      FUNCTION GetLigneDonnees(Const Row : integer;CONST ApplyMasking, FixEndianess:  BOOLEAN): TLigneDonnees;
      //Procedure GetDonnees (CONST ApplyMasking, FixEndianess:  BOOLEAN;Var MesDonnees : TDonneesEvents);
      PROCEDURE ShowTextPairs(CONST StringGrid:  TStringGrid);
      PROCEDURE WriteTextPairs(CONST Filename:  TFilename);
      PROCEDURE ShowDataRow(CONST StringGrid:  TStringGrid; CONST Row:  INTEGER;
                            CONST ApplyMasking, FixEndianess:  BOOLEAN);

      PROPERTY Version      :  STRING  READ FVersion;
      PROPERTY TextFirst    :  INTEGER READ FHeaderTextFirst;
      PROPERTY TextLast     :  INTEGER READ FHeaderTextLast;
      PROPERTY DataFirst    :  INTEGER READ FHeaderDataFirst;
      PROPERTY DataLast     :  INTEGER READ FHeaderDataLast;
      PROPERTY AnalysisFirst:  INTEGER READ FHeaderAnalysisFirst;
      PROPERTY AnalysisLast :  INTEGER READ FHeaderAnalysisLast;

      PROPERTY PROJ      :  STRING  READ FCS_PROJ_ProjectName;
      PROPERTY DATATYPE  :  STRING  READ FCS_DATATYPE;
      PROPERTY CELLS     :  STRING  READ FCS_CELLS_TypeOfCellsMeasured;
      PROPERTY Cytometer :  STRING  READ FCS_Cytometer;
      PROPERTY LaDate    :  STRING  READ FCS_LaDate;

      PROPERTY EventCount:  LongInt READ FCS_TOT_EventCount;
      PROPERTY ParmCount :  LongInt READ FCS_PAR_ParameterCount;
      PROPERTY BYTEOrder :  STRING  READ FCS_BYTEORD;

      PROPERTY Name [index:  INTEGER]:  STRING  READ GetName;
      PROPERTY NameString [index:  INTEGER]:  STRING READ GetNameString;
      PROPERTY Range[index:  INTEGER]:  INTEGER READ GetRange;
      PROPERTY Bits [index:  INTEGER]:  INTEGER READ GetBits;
      Property ParamE1[index: integer]: integer Read GetParamE1;
      Property ParamE2[index: integer]: real Read GetParamE2;
      PROPERTY Select [index:  INTEGER]:  BOOLEAN READ GetSelect WRITE SetSelect;
  END;
  function symbdec : Char;
  function NotSymbdec : Char;

IMPLEMENTATION

  USES
    SwapLibrary,dialogs;

function Symbdec : Char;
begin
  Result := DefaultFormatSettings.DecimalSeparator;// DecimalSeparator;
end;
function NotSymbdec : Char;
begin
  Result := '.';
  if DefaultFormatSettings.DecimalSeparator = Result then
    Result := ',';
end;

  CONSTRUCTOR TFCSFile.Create(CONST filename:  string);
    CONST
      HeaderSize = 58;  // bytes minimum size (assumes no user-defined segment)

    VAR
      Code           :  CHAR;
      Header         :  STRING;
      index,i          :  INTEGER;
      np             : integer;
      Keyword        :  STRING;
      Separator      :  CHAR;
      Subkey         :  STRING;
      TextData       :  STRING;
      Value          :  AnsiString;//STRING;
      Erreur : String;
  BEGIN
    FBigEndian := FALSE;    // Assume endianess same as PC (i.e., little endian)
    FCS_BYTEORD := '1,2,3,4';
    FCS_PAR_ParameterCount := 0;
    FDataElementSize := 2;  // Assume 2 bytes/element

    FFileName := filename;
    FFileStream := TFileStream.Create(FFilename, fmOpenRead);
    FFileStream.Seek(0, soFromBeginning);
    SetLength(Header, HeaderSize);  //
    FFileStream.Read(Header[1], HeaderSize);

    // Process Header
    FVersion := COPY(Header,1,6);
    Erreur := '';
    //ASSERT (COPY(Header,7,4) = '    ', 'Should be blank in header');
    if COPY(Header,7,4) <> '    ' then
      Erreur := 'L''en-Tête n''est pas conforme à la norme'+Chr(10)+chr(13)
                + 'Le logiciel ne sait pas lire ce fichier'
    else
      begin
        FHeaderTextFirst     := StrToIntDef( COPY(Header,11,8), 0);
        FHeaderTextLast      := StrToIntDef( COPY(Header,19,8), 0);
        FHeaderDataFirst     := StrToIntDef( COPY(Header,27,8), 0);
        FHeaderDataLast      := StrToIntDef( COPY(Header,35,8), 0);
        FHeaderAnalysisFirst := StrToIntDef( COPY(Header,43,8), 0);
        FHeaderAnalysisLast  := StrToIntDef( COPY(Header,51,8), 0);

        // Process Text Segment
        TextData := GetTextData;

        // First character shows separation character for parsing
        Separator := TextData[1];   // $0C for for FCS version 2
                                    // '|' for FCS version 3

        Delete(TextData, 1, 1);

        SetLength(FName ,0);
        SetLength(FNameString, 0);
        SetLength(FRange,0);
         SetLength(FMask,0);
        SetLength(FBits ,0);
        SetLength(FSelect, 0);

        WHILE LENGTH(TextData) > 0 DO
        BEGIN
          index := POS(Separator, TextData);
          Keyword := COPY(TextData, 1, index-1);
          Delete(TextData, 1, index);

          index := POS(Separator,TextData);
          IF   index = 0
          THEN BEGIN
            Value := TextData;
            TextData := ''
          END
          ELSE BEGIN
            Value := COPY(TextData,1,index-1);
            Delete(TextData,1,index);
          END;

          IF   Keyword = '$PROJ'
          THEN FCS_PROJ_ProjectName := AnsiToUTF8(Value);

          IF   Keyword = '$DATATYPE'
          THEN BEGIN
            FCS_DATATYPE := Value;
            IF   FCS_DATATYPE = 'I'
            THEN FDataElementSize := 2;

            IF   FCS_DATATYPE = 'F'     // New 15 Feb 2006
            THEN FDataElementSize := 4
          END;

          IF   Keyword = '$MODE'
          THEN FCS_MODE := Value;

          IF   Keyword = '$CELLS'
          THEN FCS_CELLS_TypeOfCellsMeasured := AnsiToUTF8(Value);

          IF   Keyword = '$CYT'
          THEN FCS_Cytometer := Value;

          IF   Keyword = '$DATE'
          THEN FCS_LaDate := Value;

          IF   Keyword = '$TOT'
          THEN FCS_TOT_EventCount := StrToInt(Trim(Value));

          IF   Keyword = '$BYTEORD'
          THEN BEGIN
            FCS_BYTEORD := Value;
            IF   (Trim(Value) = '4,3,2,1') or (Trim(Value) = '2,1,4,3')
            THEN FBigEndian := TRUE
            ELSE FCS_BYTEORD :='1,2,3,4';// ASSERT (Trim(Value) = '1,2,3,4', 'Unexpected BYTEORD ' + Value)
          END;

          IF   Keyword = '$PAR'
          THEN BEGIN
            FCS_PAR_ParameterCount := StrToInt(Trim(Value));
            SetLength(FName  ,FCS_PAR_ParameterCount);
            SetLength(FNameString,FCS_PAR_ParameterCount);
            SetLength(FRange ,FCS_PAR_ParameterCount);
            SetLength(FMask ,FCS_PAR_ParameterCount);
            SetLength(FBits  ,FCS_PAR_ParameterCount);
            SetLength(FSelect  ,FCS_PAR_ParameterCount);
            SetLength(FParamE1 ,FCS_PAR_ParameterCount);
            SetLength(FParamE2 ,FCS_PAR_ParameterCount);
          END;

          IF   COPY(Keyword,1,2) = '$P'
          THEN BEGIN
            Subkey := Keyword;
            Delete(Subkey,1,2);                  // get rid of $P
            Code := Subkey[ LENGTH(Subkey) ];
            Delete(Subkey, LENGTH(Subkey), 1);   // get rid of last letter
            index := StrToIntDef(Subkey,-1);     // index should remain

            IF   ((Code = 'N') OR (Code = 'R') OR (Code = 'B') OR (Code = 'S') OR (Code = 'E')) AND
                 (index > 0) //AND (index <= FCS_PAR_Parametercount)
            THEN BEGIN
              if index > FCS_PAR_Parametercount then //dans le cas où $PAR se trouve après
                begin
                  FCS_PAR_Parametercount := index;
                  SetLength(FName  ,FCS_PAR_ParameterCount);
                  SetLength(FNameString,FCS_PAR_ParameterCount);
                  SetLength(FRange ,FCS_PAR_ParameterCount);
                  SetLength(FMask ,FCS_PAR_ParameterCount);
                  SetLength(FBits  ,FCS_PAR_ParameterCount);
                  SetLength(FSelect  ,FCS_PAR_ParameterCount);
                  SetLength(FParamE1 ,FCS_PAR_ParameterCount);
                  SetLength(FParamE2 ,FCS_PAR_ParameterCount);
                end;
              CASE code OF
                'N':  BEGIN
                        FName[index-1] := AnsiToUTF8(value);
                        FSelect[index-1] := TRUE;  // default
                      END;
                'R':  FRange[index-1] := StrToInt(Trim(value));
                'S':  FNameString[index-1] := AnsiToUTF8(value);
                'B':  FBits[index-1]  := StrToInt(Trim(value));
                'E': begin
                       Value := Trim(Value);
                       np := pos(',',Value);
                       try
                         FParamE1[index-1] := StrToInt(Copy(Value,1,np-1));
                         Delete(Value,1,np);
                         np := pos(NotSymbdec,Value);
                         if np > 0 then
                           Value[np] := Symbdec;
                         FParamE2[index-1] := StrToFloat(Value);
                       except
                         FParamE1[index-1] := 0;
                         FParamE2[index-1] := 0;
                       end;
                     end;
              END

            END

          END;

        END;

      if (FCS_DATATYPE = 'I') and (FCS_PAR_ParameterCount > 0) then
        FDataElementSize := FBits[0] div 8;
      For i := 0 to FCS_PAR_ParameterCount - 1 do
        if FRange[i] > 0 then
          FMask[i] := FRange[i]-1
        else
          FMask[i] := $FFFF;
    end;
    // Enforce these assumptions for now

    if (FHeaderDataLast <> FHeaderDataFirst + (FCS_TOT_EventCount * FDataElementSize * FCS_PAR_ParameterCount)-1) then
      Erreur := 'Problème de déclaration du segment de données'+Chr(10)+chr(13)
                  + 'Le logiciel ne lit pas ce type de fichier';
    if (FCS_DATATYPE <> 'I') and (FCS_DATATYPE <> 'F') then
      Erreur := 'Le type de données n''est pas lisible par ce logiciel';
    if FCS_MODE <> 'L' then
      Erreur :='Le type de données n''est pas lisible par ce logiciel';
    if Erreur <>'' then
      begin
        MessageDlg(Erreur,mtError,[mbOK],0);
        FVersion := 'Erreur';
      end;
{    ASSERT(FHeaderDataLast = FHeaderDataFirst +
                             (FCS_TOT_EventCount *
                              FDataElementSize    *
                              FCS_PAR_ParameterCount)-1,
                             'Wrong assumptions about data segment');
    ASSERT((FCS_DATATYPE = 'I') (* Integer *)  OR
           (FCS_DATATYPE = 'F') (* IEEE Standard Float *),
           'Unsupported Data Type ' + FCS_DATATYPE);
    ASSERT(FCS_MODE = 'L' (* List *), 'Unsupported Mode ' + FCS_MODE);
}
  END {Create};


  DESTRUCTOR TFCSFile.Destroy;
  BEGIN
    FFileStream.Free
  END {Destroy};


  FUNCTION TFCSFile.GetName (index:  INTEGER):  STRING;
  BEGIN
    RESULT := FName[index]
  END {GetName};

  FUNCTION TFCSFile.GetNameString (index:  INTEGER):  STRING;
  BEGIN
    RESULT := FNameString[index]
  END {GetNameString};


  FUNCTION TFCSFile.GetRange(index:  INTEGER):  INTEGER;
  BEGIN
    RESULT := FRange[index]
  END {GetRange};

  FUNCTION TFCSFile.GetParamE1 (index:  Integer) : integer;
  begin
    Result := FParamE1[index];
  end; {GetParamE}

  FUNCTION TFCSFile.GetParamE2 (index:  Integer) : real;
  begin
    Result := FParamE2[index];
  end; {GetParamE}

  FUNCTION TFCSFile.GetBits (index:  INTEGER):  INTEGER;
  BEGIN
    RESULT := FBits[index]
  END {GetBits};


  FUNCTION TFCSFile.GetSelect (index:  INTEGER):  BOOLEAN;
  BEGIN
    RESULT := FSelect[index]
  END {GetSelect};

  PROCEDURE TFCSFile.SetSelect(index:  INTEGER; value:  BOOLEAN);
  BEGIN
    FSelect[index] := value
  END {SetSelect};


  FUNCTION TFCSFile.GetTextData:  STRING;
    VAR
      TextSize:  INTEGER;
  BEGIN
    TextSize := FHeaderTextLast - FHeaderTextFirst;
    SetLength(RESULT, TextSize);
    FFileStream.Seek(FHeaderTextFirst, soFromBeginning);
    FFileStream.Read(RESULT[1], TextSize);
  END;


  PROCEDURE TFCSFile.ShowTextPairs(CONST StringGrid:  TStringGrid);
    VAR
      index    :  INTEGER;
      Separator:  CHAR;
      TextData :  STRING;
  BEGIN
    TextData := GetTextData;

    // First character shows separation character for parsing
    Separator := TextData[1];   // '|' for FCS versions 1 and 2
                                // '/' for FCS version 3

    Delete(TextData, 1, 1);

    WHILE LENGTH(TextData) > 0 DO
    BEGIN
      index := POS(Separator, TextData);
      StringGrid.Cells[1, StringGrid.RowCount-1] := COPY(TextData, 1, index-1);
      Delete(TextData, 1, index);

      index := POS(Separator,TextData);
      IF   index = 0
      THEN BEGIN
        StringGrid.Cells[2, StringGrid.RowCount-1] := TextData;
        TextData := ''
      END
      ELSE BEGIN
        StringGrid.Cells[2, StringGrid.RowCount-1] := COPY(TextData,1,index-1);
        Delete(TextData,1,index);
      END;

      IF   LENGTH(TextData) > 0
      THEN StringGrid.RowCount := StringGrid.RowCount + 1;
    END

  END {ShowTextPairs};


  PROCEDURE TFCSFile.WriteTextPairs(CONST Filename:  TFilename);
    VAR
      header   :  TextFile;
      index    :  INTEGER;
      keyword  :  STRING;
      Separator:  CHAR;
      TextData :  STRING;
      Value    :  STRING;
  BEGIN
    AssignFile(header, Filename);
    REWRITE(header);
    TRY
      WRITELN(header, '"Keyword","Value"');
      WRITELN(header, '"Version","', FVersion, '"');

      TextData := GetTextData;

      // First character shows separation character for parsing
      Separator := TextData[1];   // '|' for FCS versions 1 and 2
                                  // '/' for FCS version 3

      Delete(TextData, 1, 1);

      WHILE LENGTH(TextData) > 0 DO
      BEGIN
        index := POS(Separator, TextData);
        Keyword := COPY(TextData, 1, index-1);
        Delete(TextData, 1, index);

        index := POS(Separator,TextData);
        IF   index = 0
        THEN BEGIN
          Value := TextData;
          TextData := ''
        END
        ELSE BEGIN
          Value := COPY(TextData,1,index-1);
          Delete(TextData,1,index);
        END;
        WRITELN(header, Keyword, ',', Value);
      END

    FINALLY
      CloseFile(header)
    END
  END {WriteTextPairs};


  // NameFormat = 0 for "short" names, 1 for "long" names, and
  // 2 for short-long names
  FUNCTION  TFCSFile.GetNameRow(CONST NameFormat:  INTEGER):  STRING;
    CONST
      ElementSize = 2;

    VAR
      i        :  INTEGER;
      name     :  STRING;

  BEGIN
    // HeaderRow
    RESULT := '';
    FOR i := 0 TO FCS_PAR_ParameterCount-1 DO
    BEGIN
      IF   FSelect[i]
      THEN
        CASE NameFormat OF
          0:  name := GetName(i);
          1:  name := GetNameString(i);
          ELSE
            IF  GetName(i) = GetNameString(i)
            THEN name := GetName(i)
            ELSE name := GetName(i) + '-' + GetNameString(i)
        END;

        IF   length(name) = 0
        THEN name := 'F' + IntToStr(i);

      RESULT := RESULT + '"' + name + '"' + ',';
    END;

    Delete(RESULT, LENGTH(RESULT), 1);  // get rid of final ','
  END {GetNameRow};
  FUNCTION TFCSFile.GetLigneDonnees(Const Row : integer;CONST ApplyMasking, FixEndianess:  BOOLEAN): TLigneDonnees;
    Var
      ElementSize:  INTEGER;
      i          :  INTEGER;
      RowStart   :  INTEGER;
      Value      :  WORD;
      Value4     :  LongWord;
      ValeurB    :  Byte;
      ShortFloat :  SINGLE  ABSOLUTE Value4;  // Memory overlay trick
  begin
    ElementSize := FDataElementSize;  // 2 for 'I' or 4 for 'F'
    SetLength(Result,FCS_PAR_ParameterCount);
    // "Normal" data rows
    RowStart := FHeaderDataFirst + (Row - 1) * ElementSize * FCS_PAR_ParameterCount;

    // Caution:  Indexing is a bit confusing here
    FOR i := 1 TO FCS_PAR_ParameterCount DO
      BEGIN
        FFileStream.Seek(RowStart+ElementSize*(i-1), soFromBeginning);
        IF   FCS_DATATYPE = 'I' THEN
          BEGIN
            if ElementSize = 1 then
              begin
                FFileStream.Read(ValeurB,1);
                RESULT[i-1] := ValeurB;
              end
            else if ElementSize = 2 then
              begin
                FFileStream.Read(Value,2);
                if FBigEndian then
                  Value:=Swap(Value);
                Value := Value AND FMask[i-1];
                RESULT[i-1] := Value;
              end
            else if ElementSize = 4 then
              begin
                FFileStream.Read(Value4,4);
                if FBigEndian then
                  Value4:=Swap(Value4);
                Value4 := Value4 AND FMask[i-1];
                RESULT[i-1] := Value4;
              end;
            // Mask off extra bits that should be 0 but are not for some reason.
            // For example if Range is 4096, the mask = 4095 = $0FFF;
            // if Range is 1024, the mask = 1023 = $03FF
          END;

        IF FCS_DATATYPE = 'F' THEN
          BEGIN
            FFileStream.Read(Value4, ElementSize);

            IF FBigEndian THEN
              Value4 := Swap4(Value4);

            RESULT[i-1] := ShortFloat;
            if Result[i-1] < 0 then
              Result[i-1] := 0;
          END;
    END;

  end;

  FUNCTION  TFCSFile.GetDataRow(CONST Row:  INTEGER;
                                CONST ApplyMasking, FixEndianess:  BOOLEAN):  STRING;
    VAR
      ElementSize:  INTEGER;
      i          :  INTEGER;
      RowStart   :  INTEGER;
      Value      :  WORD;
      Value4     :  LongWord;
      ShortFloat :  SINGLE  ABSOLUTE Value4;  // Memory overlay trick
  BEGIN
    ASSERT(Row > 0);

    ElementSize := FDataElementSize;  // 2 for 'I' or 4 for 'F'

    // "Normal" data rows
    RowStart := FHeaderDataFirst + (Row - 1) * ElementSize * FCS_PAR_ParameterCount;

    RESULT := '';
    // Caution:  Indexing is a bit confusing here
    FOR i := 1 TO FCS_PAR_ParameterCount DO
    BEGIN
      IF   FSelect[i-1]
      THEN BEGIN
        FFileStream.Seek(RowStart+ElementSize*(i-1), soFromBeginning);

        IF   FCS_DATATYPE = 'I'
        THEN BEGIN

          FFileStream.Read(Value, ElementSize);

          IF   FBigEndian
          THEN Value := Swap2(Value);

          // Mask off extra bits that should be 0 but are not for some reason.
          // For example if Range is 4096, the mask = 4095 = $0FFF;
          // if Range is 1024, the mask = 1023 = $03FF
          IF   FRange[i-1] > 0
          THEN Value := Value AND (FRange[i-1] - 1);

          RESULT := RESULT + IntToStr(Value) + ',';
        END;

        IF   FCS_DATATYPE = 'F'
        THEN BEGIN
          FFileStream.Read(Value4, ElementSize);

          IF   FBigEndian
          THEN Value4 := Swap4(Value4);

          RESULT := RESULT + Format('%f', [ShortFloat]) + ','
        END;

      END
    END;

    Delete(RESULT, LENGTH(RESULT), 1);  // get rid of final ','
  END {GetDataRow};


  // Row is 1-origin
  PROCEDURE TFCSFile.ShowDataRow(CONST StringGrid:  TStringGrid;
                                 CONST Row:  INTEGER;
                                 CONST ApplyMasking, FixEndianess:  BOOLEAN);
    CONST
      Offset      = 4;

    VAR
      ElementSize:  INTEGER;
      i          :  INTEGER;
      RowStart   :  INTEGER;
      Value      :  WORD;
      Value4     :  LongWord;
      ShortFloat :  SINGLE  ABSOLUTE Value4;  // Memory overlay trick
  BEGIN
    ElementSize := FDataElementSize;  // 2 for 'I' or 4 for 'F'

    RowStart := FHeaderDataFirst + (Row - 1) * ElementSize * FCS_PAR_ParameterCount;

    StringGrid.Cells[0, Offset+Row] := IntToStr(Row);

    FOR i := 1 TO FCS_PAR_ParameterCount DO
    BEGIN
      FFileStream.Seek(RowStart+ElementSize*(i-1), soFromBeginning);

      IF   FCS_DATATYPE = 'I'
      THEN BEGIN

        FFileStream.Read(Value, ElementSize);

        IF   FBigEndian
        THEN Value := Swap2(Value);

        // Mask off extra bits that should be 0 but are not for some reason.
        // For example if Range is 4096, the mask = 4095 = $0FFF;
        // if Range is 1024, the mask = 1023 = $03FF
        IF   FRange[i-1] > 0
        THEN Value := VAlue AND (FRange[i-1] - 1);

        StringGrid.Cells[i, Offset+Row] := IntToStr(Value)
      END;

      IF   FCS_DATATYPE = 'F'
      THEN BEGIN

        FFileStream.Read(Value4, ElementSize);

        IF   FBigEndian
        THEN Value4 := Swap4(Value4);

        StringGrid.Cells[i, Offset+Row] := Format('%f', [ShortFloat])
      END;


    END
  END {ShowDataRow};


END.
