unit FormStats;
{$mode delphi}
interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Dialogs, ExtCtrls, Menus, StdCtrls,Donnees;

type

  { TFStats }

  TFStats = class(TForm)
    MemoStats: TMemo;
    MainMenu1: TMainMenu;
    MAffichage: TMenuItem;
    MGraphique: TMenuItem;
    MSurDessus: TMenuItem;
    MFermer: TMenuItem;
    MEdition: TMenuItem;
    MToutCopier: TMenuItem;
    MCopierSelection: TMenuItem;
    MFichier: TMenuItem;
    MEnregistrer: TMenuItem;
    SaveStats: TSaveDialog;
    procedure MGraphiqueClick(Sender: TObject);
    procedure MSurDessusClick(Sender: TObject);
    procedure MEnregistrerClick(Sender: TObject);
    procedure MFermerClick(Sender: TObject);
    procedure MToutCopierClick(Sender: TObject);
    procedure MCopierSelectionClick(Sender: TObject);
  private
    { Déclarations privées }
  public
    { Déclarations publiques }
    CatRestr : TCategorie;
    procedure AfficheStats;
    function Afficher(LaCategorie : TCategorie) : Boolean;
    // utiliser dans LesDonnees[i]
{     MoysP : Array of real; // pour stocker les moyennes pour chaque paramètre
      EcTsP : Array of real; // pour stocker les ecart-types pour chaque paramètre
      MoysPCS : Array of Array of real; // pour stocker les moyennes des paramètres de chaque catégorie [i,j] i : categorie, j : paramètre
      EcTsPCS : Array of Array of real; // pour stocker les ecart-types des paramètres de chaque catégorie [i,j] i : categorie, j : paramètre
      NbEventCS : Array of real;
      MoysPCC : Array of Array of real; // pour stocker les moyennes des paramètres de chaque catégorie [i,j] i : categorie, j : paramètre
      EcTsPCC : Array of Array of real; // pour stocker les ecart-types des paramètres de chaque catégorie [i,j] i : categorie, j : paramètre
      NbEventCC : Array of real;
}
  end;

var
  FStats: TFStats;

implementation

{$R *.lfm}
 uses math,formtableau;
function TFStats.Afficher(LaCategorie : TCategorie) : Boolean;
Var
  i : integer;
begin
  Result := (CatRestr = nil);
  if CatRestr <> nil then
    begin
      if LaCategorie.Composite then
        begin
          for i := 0 to Length(LaCategorie.CatAscendants)-1 do
            if LaCategorie.CatAscendants[i] = CatRestr then
              begin
                Result := True;
                Break;
              end;
        end;
    end;
end;

procedure TFStats.AfficheStats;
Var
  i,j,k,n : integer;
  ligne : string;
  ValLim1,ValLim2 : integer;
  LaCat : TCategorie;
  NomPar1,NomPar2 : string;
begin
  MemoStats.Lines.Clear;
  MemoStats.Lines.Add('Statistiques');
  for i := 0 to Length(LesDonneesCyto) - 1 do
    begin
      if LesDonneesCyto[i].NbEvenements > 0 then
        begin
          if LesDonneesCyto[i].Projet <> '' then
            MemoStats.Lines.Add('Nom du projet : '+ LesDonneesCyto[i].Projet);
          if LesDonneesCyto[i].Cellules <> '' then
            MemoStats.Lines.Add('Cellules étudiées : '+ LesDonneesCyto[i].Cellules);

          LesDonneesCyto[i].CalculeStats;
          MemoStats.Lines.Add('');
          MemoStats.Lines.Add('Série ' + LesDonneesCyto[i].Nom);
          MemoStats.Lines.Add('  Donnees générales : '+ IntToStr(LesDonneesCyto[i].NbEvenements) + ' événements (cellules)');
          for j := 0 to LesDonneesCyto[i].NbParam - 1 do
            begin
              ligne := '    ' + LesDonneesCyto[i].Parametres[j].Alias;
              ligne := ligne + ' : moyenne ' + FloatToStrF(LesDonneesCyto[i].MoysP[j],ffFixed,8,2);
              ligne := ligne + ' : écart type ' + FloatToStrF(LesDonneesCyto[i].EcTsP[j],ffFixed,8,2);
              MemoStats.Lines.Add(ligne);
            end;
          if (Catrestr <> nil) and Compatible(Catrestr,LesDonneesCyto[i])  then
            begin
              ligne := 'Restriction à la catégorie '+ Catrestr.Nom;
              LaCat := Catrestr;
              j := LaCat.Numero;
              if LaCat.NbParams = 1 then
                begin
                  ValLim1 := Round(LaCat.ValeurMin);
                  ValLim2 := Round(LaCat.ValeurMax);
                  NomPar1 := Trim(LaCat.Params[0].Alias);
                  if NomPar1 = '' then
                    NomPar1 := LaCat.Params[0].Nom;
                  if ValLim2 < LaCat.Params[0].Resolution then
                    begin
                      if LaCat.Params[0].Puis10Log >0 then  // on tient compte de Puis10log
                        begin
                          ValLim1 := Round(power(10,ValLim1/LaCat.Params[0].Resolution*LaCat.Params[0].Puis10Log));
                          ValLim2 := Round(power(10,ValLim2/LaCat.Params[0].Resolution*LaCat.Params[0].Puis10Log));
                        end;
                      ligne := ligne + ' définie par des valeurs de ' + NomPar1
                                     + ' comprises entre ' + IntToStr(ValLim1) + ' et '+ IntToStr(ValLim2);
                    end
                  else
                    begin
                      if LaCat.Params[0].Puis10Log >0 then // on tient compte de Puis10log
                        ValLim1 := Round(power(10,ValLim1/LaCat.Params[0].Resolution*LaCat.Params[0].Puis10Log));

                      ligne := ligne + ' définie par des valeurs de ' + NomPar1
                                     + ' supérieures à ' + IntToStr(ValLim1);
                    end;
                end
              else if LaCat.NbParams = 2 then
                begin
                  NomPar1 := Trim(LaCat.Params[0].Alias);
                  if NomPar1 = '' then
                    NomPar1 := LaCat.Params[0].Nom;
                  NomPar2 := LaCat.Params[1].Alias;
                  if NomPar2 = '' then
                    NomPar2 := LaCat.Params[1].Nom;

                  ligne := ligne + ' définie par un polygone dans le graphe représentant '+ NomPar1
                             + ' en fonction de '+ NomPar2;
                end;
              MemoStats.Lines.Add(ligne);
              ligne := '      nombre : ' + Inttostr(LesDonneesCyto[i].NbEventCS[j]);
              ligne := ligne +' soit '+ FloatToStrF(100*(LesDonneesCyto[i].NbEventCS[j]/LesDonneesCyto[i].NbEvenements),ffFixed,8,2) +' % du total';
              MemoStats.Lines.Add(ligne);
              for k := 0 to LesDonneesCyto[i].NbParam - 1 do
                begin
                  ligne := '      ' + LesDonneesCyto[i].Parametres[k].Alias;
                  ligne := ligne + ' : moyenne ' + FloatToStrF(LesDonneesCyto[i].MoysPCS[j,k],ffFixed,8,2);
                  ligne := ligne + ' : écart type ' + FloatToStrF(LesDonneesCyto[i].EcTsPCS[j,k],ffFixed,8,2);
                  MemoStats.Lines.Add(ligne);
                end;
            end
          else if LesCategories.NbCategorieDeBase > 0 then
            begin
              MemoStats.Lines.Add('  Donnees par catégories : ');
              For j := 0 to LesCategories.NbCategorieDeBase - 1 do
                begin
                  LaCat := LesCategories.LesCategoriesDeBase[j];
                  if Compatible(LaCat,LesDonneesCyto[i]) then
                    begin
                      ligne :='    Catégorie ' + LaCat.Nom;
                      if LaCat.NbParams = 1 then
                        begin
                          ValLim1 := Round(LaCat.ValeurMin);
                          ValLim2 := Round(LaCat.ValeurMax);
                          NomPar1 := Trim(LaCat.Params[0].Alias);
                          if NomPar1 = '' then
                            NomPar1 := LaCat.Params[0].Nom;
                          if ValLim2 < LaCat.Params[0].Resolution then
                            begin
                              if LaCat.Params[0].Puis10Log >0 then  // on tient compte de Puis10log
                                begin
                                  ValLim1 := Round(power(10,ValLim1/LaCat.Params[0].Resolution*LaCat.Params[0].Puis10Log));
                                  ValLim2 := Round(power(10,ValLim2/LaCat.Params[0].Resolution*LaCat.Params[0].Puis10Log));
                                end;
                              ligne := ligne + ' définie par des valeurs de ' + NomPar1
                                             + ' comprises entre ' + IntToStr(ValLim1) + ' et '+ IntToStr(ValLim2);
                            end
                          else
                            begin
                              if LaCat.Params[0].Puis10Log >0 then // on tient compte de Puis10log
                                ValLim1 := Round(power(10,ValLim1/LaCat.Params[0].Resolution*LaCat.Params[0].Puis10Log));

                              ligne := ligne + ' définie par des valeurs de ' + NomPar1
                                             + ' supérieures à ' + IntToStr(ValLim1);
                            end;
                        end
                      else
                        begin
                          NomPar1 := Trim(LaCat.Params[0].Alias);
                          if NomPar1 = '' then
                            NomPar1 := LaCat.Params[0].Nom;
                          NomPar2 := LaCat.Params[1].Alias;
                          if NomPar2 = '' then
                            NomPar2 := LaCat.Params[1].Nom;

                          ligne := ligne + ' définie par un polygone dans le graphe représentant '+ NomPar1
                                     + ' en fonction de '+ NomPar2;
                        end;
                      MemoStats.Lines.Add(ligne);
                      ligne := '      nombre : ' + Inttostr(LesDonneesCyto[i].NbEventCS[j]);
                      ligne := ligne +' soit '+ FloatToStrF(100*(LesDonneesCyto[i].NbEventCS[j]/LesDonneesCyto[i].NbEvenements),ffFixed,8,2) +' % du total';
                      MemoStats.Lines.Add(ligne);
                      for k := 0 to LesDonneesCyto[i].NbParam - 1 do
                        begin
                          ligne := '      ' + LesDonneesCyto[i].Parametres[k].Alias;
                          ligne := ligne + ' : moyenne ' + FloatToStrF(LesDonneesCyto[i].MoysPCS[j,k],ffFixed,8,2);
                          ligne := ligne + ' : écart type ' + FloatToStrF(LesDonneesCyto[i].EcTsPCS[j,k],ffFixed,8,2);
                          MemoStats.Lines.Add(ligne);
                        end;
                    end;
                  //if Not LaCat.CatAnnexe then
                  //if LaCat.Active then
                    //begin

                    //end;
                end;
            end;
          For j := 0 to LesCategories.NbCategorieComp - 1 do   //** à rattacher éventuellement aux catégories d'origine
            begin
              LaCat := LesCategories.LesCategoriesComp[j];
              if Compatible(LaCat,LesDonneesCyto[i]) and Afficher(LaCat) then
                begin
                  ligne :='    Catégorie ' + LaCat.Nom;
                  if LaCat.Lien = tlEt then
                    begin
                      ligne := ligne +' formée des cellules appartenant à la fois aux '+ LaCat.CatParents[0].Nom
                                     +' et aux '+ LaCat.CatParents[1].Nom;
                    end
                  else if LaCat.Lien = tlEt then
                    ligne := ligne+' qui réunit les catégories '+ LaCat.CatParents[0].Nom
                                  +' et '+LaCat.CatParents[1].Nom;

                  MemoStats.Lines.Add(ligne);
                  ligne := '       nombre : ' + Inttostr(LesDonneesCyto[i].NbEventCC[j]);
                  if LaCat.Lien = tlEt then
                    begin
                      if LesDonneesCyto[i].NbEventCC[j]> 0 then
                        begin
                          ligne := ligne +' soit '+ FloatToStrF(100*(LesDonneesCyto[i].NbEventCC[j]/LesDonneesCyto[i].NbEvenements),ffFixed,8,2) +' % du total';
                          for k := 0 to 1 do
                            if LesDonneesCyto[i].NbEventCS[LesCategories.LesCategoriesComp[j].CatParents[k].Numero] > 0 then
                              ligne := ligne + ' ; ' + FloatToStrF(100*(LesDonneesCyto[i].NbEventCC[j]/LesDonneesCyto[i].NbEventCS[LesCategories.LesCategoriesComp[j].CatParents[k].Numero]),ffFixed,8,2)
                                + ' % des '+ LaCat.CatParents[k].Nom;
                        end;
                    end;
                  MemoStats.Lines.Add(ligne);
                    for k := 0 to LesDonneesCyto[i].NbParam - 1 do
                      begin
                        ligne := '      ' + LesDonneesCyto[i].Parametres[k].Alias;
                        ligne := ligne + ' : moyenne ' + FloatToStrF(LesDonneesCyto[i].MoysPCC[j,k],ffFixed,8,2);
                        ligne := ligne + ' : écart type ' + FloatToStrF(LesDonneesCyto[i].EcTsPCC[j,k],ffFixed,8,2);
                        MemoStats.Lines.Add(ligne);
                      end;
                end;
            end;
        end;
    end;
  n := 0;
  if CatRestr <> nil then
    begin
      for i := 0 to LesCategories.NbCategorieComp -1 do
        if Afficher(LesCategories.LesCategoriesComp[i]) then
          inc(n);
    end
  else if LesCategories.NbCategorieDeBase > 0 then
    inc(n);
  if (n > 0) and (Length(LesDonneesCyto) >2) and (Trim(LesDonneesCyto[0].ParamSerie)<>'') then
    MGraphique.Visible:=True
  else
    MGraphique.Visible:=False;
end;
procedure TFStats.MCopierSelectionClick(Sender: TObject);
Var
  MonMessage : string;
begin
  if MemoStats.SelLength > 0 then
    MemoStats.CopyToClipboard
  else
    begin
      MonMessage := 'Rien n''est sélectionné';
      MessageDlg(MonMessage,TMsgDlgType.mtError,[TMsgDlgBtn.mbOK],0);
    end;

end;

procedure TFStats.MEnregistrerClick(Sender: TObject);
begin
  if MemoStats.Lines.Count > 0 then
    begin
      if SaveStats.Execute then
        MemoStats.Lines.SaveToFile(SaveStats.FileName);
    end;
end;

procedure TFStats.MSurDessusClick(Sender: TObject);
begin
  MSurDessus.Checked:=Not MSurDessus.Checked;
  if MSurDessus.Checked then
    FormStyle := fsStayOnTop
  else
    FormStyle := fsNormal;
end;

procedure TFStats.MGraphiqueClick(Sender: TObject);
Var
  UnMenu : TMenuItem;
  i, n : integer;
begin
  FTableau.MCategories.Clear;
  n := 0;
  if CatRestr <> nil then
    begin
      FTableau.numCatRestr:=LesCategories.LeNumero(CatRestr);
      for i := 0 to LesCategories.NbCategorieComp -1 do
        if Afficher(LesCategories.LesCategoriesComp[i]) then
          begin
            UnMenu := TMenuItem.Create(FTableau.MCategories);
            UnMenu.Caption:=LesCategories.LesCategoriesComp[i].Nom;
            UnMenu.Tag:=-2-i;
            UnMenu.OnClick :=FTableau.MCategorieClick;
            FTableau.MCategories.Add(UnMenu);
            inc(n);
          end;
    end
  else if LesCategories.NbCategorieDeBase > 0 then
    begin
      FTableau.numCatRestr:=-1;
      for i := 0 to LesCategories.NbCategorieDeBase -1 do
        begin
          UnMenu := TMenuItem.Create(FTableau.MCategories);
          UnMenu.Caption:=LesCategories.LesCategoriesDeBase[i].Nom;
          UnMenu.Tag:=i;
          UnMenu.OnClick :=FTableau.MCategorieClick;
          FTableau.MCategories.Add(UnMenu);
          inc(n);
        end;
      for i := 0 to LesCategories.NbCategorieComp -1 do
        begin
          UnMenu := TMenuItem.Create(FTableau.MCategories);
          UnMenu.Caption:=LesCategories.LesCategoriesComp[i].Nom;
          UnMenu.Tag:=-2-i;
          UnMenu.OnClick :=FTableau.MCategorieClick;
          FTableau.MCategories.Add(UnMenu);
          inc(n);
        end;
    end;
  if (n > 0) and (Length(LesDonneesCyto) >2) then
    begin
      FTableau.MCategories.Items[0].Checked:=True;
      FTableau.Afficher;
      FTableau.Show;
      if FTableau.WindowState=wsMinimized then
        FTableau.WindowState:=wsNormal;
      FTableau.BringToFront;
      Application.ProcessMessages;
    end;
end;

procedure TFStats.MFermerClick(Sender: TObject);
begin
  Close;
end;

procedure TFStats.MToutCopierClick(Sender: TObject);
begin
  MemoStats.SelectAll;
  MemoStats.CopyToClipboard;
  MemoStats.SelStart := 0;
  MemoStats.SelLength :=0;
end;

end.
