// efg, Stowers Institute, 15 Aug 2003. Modified 27 Apr 2004.
//
// Modified 15 February 2006 to support Float $DataType used
// Utilisation de parties de FCSLibrary.pas dont l'en-tête est reproduite ci-desssous
// with BD (Becton, Dickinson) FACSDiva Software for
// See http://facs.stanford.edu/sff/doc/bdfacsdivasoftwareuserguide.pdf
//
// Released under GNU General Public License
//
// References:
//
// Data File Standard for Flow Cyometry, Version FCS 3.0
//      http://greenhorn.stc.cmu.edu/FCSAPI/FCS3.html
//
// Seamer, L.C, et al, Proposed New Data File Standard for Flow Cytometry,
//      Version FCS 3.0, Cytometry, 28:118-122, 1997.
//      http://greenhorn.stc.cmu.edu/publications/64-seamer1997.pdf
//
// Leif, Robert C and Suzanne B. Leif, The Evolution of Flow Cytometry
//      Standard, FCS3.0, into a DICOM Compatible Format, Optical
//      Diagnostics of Biological Fluids and Advanced Techniques in
//      Analytical Cytogy, SPIE Procs Series, Vol 2982, pp. 1-13, 1997.
//      http://newportinstruments.com/cytometryml/fcs_to_dicom1.pdf
//
// *** Modifications Jean-françois Madre du 19/02/13 au 26/03/13 pour adapter au logiciel Cytométrie
// *** modifs pour lire quelques paramètres supplémentaires et pour la lecture des fichiers avec des données de type Entier (I)
//      lorsque le codage n'est pas sur 2 bits
// *** remplacement des Assert par des MessageDLg pour écrire un message si le format de fichier n'est pas lu
// *** Modification pour lire des fichiers FCS2.0 dont $PAR est après les $PnN...
// *** Modification pour lire les fichiers comportant plusieurs jeux de données

UNIT gestionFichier;

{$MODE Delphi}

INTERFACE

USES
  Classes,     // TFileStream
  //Grids,       // TStringGrid
  SysUtils,    // TFileName
  Donnees, //TDonneesCyto
  basis,
  LU;
TYPE
  TLigneDonnees = Array of Real;
  TFCSFile = Class(TObject)
      FFileName : TFileName;
      FFileStream : TFileStream;
      OffsetGen : int64; //Offset entre le début et le début du jeu de données en cours de lecture
      // paramètres de chaque jeu de données
      // variables de l'en-tête générale du jeu
      FHeaderTextFirst    :  INTEGER;   // Bytes 10-17  (0-based)
      FHeaderTextLast     :  INTEGER;   // Bytes 18-25
      FHeaderDataFirst    :  INTEGER;   // Bytes 26-33
      FHeaderDataLast     :  INTEGER;   // Bytes 34-41
      FHeaderAnalysisFirst :  INTEGER;   // Bytes 42-49
      FHeaderAnalysisLast :  INTEGER;   // Bytes 50-57
      Version : string;
      SpillMatrix : Array of Array of real;
      CMatrix : Array of Array of real;//matrice de correction
      InvSpillMatrix : Array of Array of real;
      ParamCompN : Array of string;
      FDataElementSize : integer;   // 2 for 'I' or 4 for 'F' DATATYPE

      FPROJ : string;
      FDATATYPE : string;
      CELLS : string;
      FMODE : string;
      EventCount : LongInt;
      ParameterCount : LongInt;

      BYTEORD : string;   // par déf aut:  1,2,3,4 (PC)
      FBigEndian : Boolean;  // supposé FALSE

      Cytometer : string;
      LaDate : string;
      // variables des paramètres
      FName : array of string;
      FNameString : array of string;
      FRange : array of integer;
      FMask : array of integer;
      FBits : array of integer;
      FParamE1 : array of integer; // ajout JFM
      FParamE2 : array of real; // ajout JFM
      FNextData : int64;
      FFile : string;
      FNomParamSerie : string;
      FUniteParamSerie : string;
      FValeurParamSerie : integer;

      Constructor Create(Const Filename : string);
      Destructor  Destroy; Override;
      function LitEnTete : Boolean;
      function LitTextData : Boolean;
      function GetTextData : string;
      function LitJeuDonnees : Boolean;
      function GetLigneDonnees(Const Row : integer;CONST ApplyMasking, FixEndianess:  BOOLEAN): TLigneDonnees;
      procedure Inverse_LU;
      //procedure InvertSpillMatrix;
     { procedure addrows(r1,r2: integer; destrow: integer);
      procedure multrow(adouble: real; destrow: integer);
      procedure solvecell(arow,acol: integer; desired: real);
      procedure populatediagonal(arow: integer);
      procedure solveleft(arow: integer);
      procedure solveright(arow: integer);
      }
      PROPERTY TextFirst : INTEGER READ FHeaderTextFirst;
      PROPERTY TextLast     :  INTEGER READ FHeaderTextLast;
      PROPERTY DataFirst    :  INTEGER READ FHeaderDataFirst;
      PROPERTY DataLast     :  INTEGER READ FHeaderDataLast;
      PROPERTY AnalysisFirst:  INTEGER READ FHeaderAnalysisFirst;
      PROPERTY AnalysisLast :  INTEGER READ FHeaderAnalysisLast;
  END;
Var
  OrdreBitsDefaut : string;
  InverserBits : Boolean;
  function symbdec : Char;
  function NotSymbdec : Char;
  procedure ChangeOrdreBitsDefaut(inverse : Boolean);
  //function isdiff(a,b: real): boolean;
Const
  epsilon=1e-10;

IMPLEMENTATION

  USES
    SwapLibrary,dialogs,forms,cyto,fmatrice;

procedure ChangeOrdreBitsDefaut(inverse : Boolean);
begin
  InverserBits := inverse;
  if InverserBits then
    OrdreBitsDefaut :='4,3,2,1'
  else
    OrdreBitsDefaut :='1,2,3,4';
end;

function Symbdec : Char;
begin
  Result := DefaultFormatSettings.DecimalSeparator;// DecimalSeparator;
end;
function NotSymbdec : Char;
begin
  Result := '.';
  if DefaultFormatSettings.DecimalSeparator = Result then
    Result := ',';
end;


function TFCSFile.LitEnTete : Boolean;
Const
  HeaderSize = 58;  // bytes minimum size (assumes no user-defined segment)
Var
  Header : string;
  Erreur : String;
begin
  Result := False;
  SetLength(Header, HeaderSize);
  FFileStream.Seek(OffsetGen,soFromBeginning);
  FFileStream.Read(Header[1], HeaderSize);

  // Process Header
  Version := COPY(Header,1,6);
  Erreur := '';
  //ASSERT (COPY(Header,7,4) = '    ', 'Should be blank in header');
  if COPY(Header,7,4) <> '    ' then
    Erreur := 'L''en-Tête n''est pas conforme à la norme'+Chr(10)+chr(13)
              + 'Le logiciel ne sait pas lire ce fichier'
  else
    begin
      FHeaderTextFirst     := StrToIntDef( COPY(Header,11,8), 0);
      FHeaderTextLast      := StrToIntDef( COPY(Header,19,8), 0);
      FHeaderDataFirst     := StrToIntDef( COPY(Header,27,8), 0);
      FHeaderDataLast      := StrToIntDef( COPY(Header,35,8), 0);
      FHeaderAnalysisFirst := StrToIntDef( COPY(Header,43,8), 0);
      FHeaderAnalysisLast  := StrToIntDef( COPY(Header,51,8), 0);
      //FNextData := FHeaderTextFirst;
      Result := True;
   end;
end;
function TFCSFile.LitJeuDonnees : Boolean;
Var
  Code :Char;
  index,i,j : integer;
  np,n : integer;
  Keyword : string;
  Separator : Char;
  Subkey : string;
  TextData : string;
  Value :  AnsiString;//STRING;
  Erreur : String;
  LaDonneeCyto : TDonneesCyto;
  LaLigneDonnees : TLigneDonnees;
  UnParam : TParam;
  OnePercent : integer;
begin
  Result := False;
  if LitEnTete then
    begin
      if LitTextData then
        begin
          FCyto.ProgressBar.Position := 0;
          FCyto.ProgressBar.Visible := True;
          Application.ProcessMessages;
          OnePercent := EventCount div 100;
          n := Length(LesDonneesCyto);
          SetLength(LesDonneesCyto,n+1);
          LaDonneeCyto := TDonneesCyto.Create;
          LesDonneesCyto[n] := LaDonneeCyto;
          LaDonneeCyto.Nom := FFile;
          LaDonneeCyto.Version := Version;
          LaDonneeCyto.Cellules:= CELLS;
          LaDonneeCyto.Projet := FProj;
          LaDonneeCyto.OrdreBits := BYTEOrd;
          LaDonneeCyto.LaDate := LaDate;
          LaDonneeCyto.Cytometre := Cytometer;
          LaDonneeCyto.NbParam:=ParameterCount;
          LaDonneeCyto.TypeDonnees:=FDATATYPE;
          LaDonneeCyto.ParamSerie:=FNomParamSerie;
          LaDonneeCyto.UniteParamSerie:=FUniteParamSerie;
          LaDonneeCyto.ValeurParamSerie:=FValeurParamSerie;
          SetLength(LaDonneeCyto.Parametres,LaDonneeCyto.NbParam);
          SetLength(LaDonneeCyto.Maxi,LaDonneeCyto.NbParam);
          SetLength(LaDonneeCyto.Maxi2,LaDonneeCyto.NbParam);
          for i := 0 to LaDonneeCyto.NbParam-1 do
            begin
              UnParam := TParam.Create;
              UnParam.Numero := i;
              UnParam.Nom := FName[i];
              UnParam.Alias :=FNameString[i];//LitParamT('$P'+IntToStr(i)+'S');
              if (UnParam.Alias = '') and (UpperCase(UnParam.Nom) = 'TIME') then
                UnParam.Alias := 'Temps';
              UnParam.Nombrebits := FBits[i];//LitParamN('$P'+IntToStr(i)+'B');
              UnParam.Puis10Log := FParamE1[i];//LitParamR('$P'+IntToStr(i)+'E');
              UnParam.ValPourZero:=FParamE2[i]; //** inutilisé sauf pour la sauvegarde
              if UnParam.Puis10Log > 0 then
                UnParam.TypeEchelle := teLog
               else
                 UnParam.TypeEchelle := teLin;
               UnParam.Resolution := FRange[i];//LitParamN('$P'+IntToStr(i)+'R');
               UnParam.Masque:=UnParam.Resolution - 1;
               LaDonneeCyto.Parametres[i] := UnParam;
               LaDonneeCyto.Resolution := UnParam.Resolution;
               LaDonneeCyto.Maxi[i] := 0;
               LaDonneeCyto.Maxi[i] := 0;
            end;

          LaDonneeCyto.NbEvenements:=EventCount;
          SetLength(LaDonneeCyto.Donnees,LaDonneeCyto.NbEvenements);
          SetLength(LaDonneeCyto.Couleurs,LaDonneeCyto.NbEvenements);

          for i := 0 to EventCount - 1 do
            begin
              SetLength(LaDonneeCyto.Donnees[i],LaDonneeCyto.NbParam);
              if i mod OnePercent = 0 then
                begin
                  FCyto.ProgressBar.Position := Round(100 * i / EventCount);
                  Application.ProcessMessages;
                end;
              LaLigneDonnees:=GetLigneDonnees(i+1,True,True);
              for j := 0 to ParameterCount -1 do
                begin
                  LaDonneeCyto.Donnees[i,j]:=LaLigneDonnees[j];
                  if LaLigneDonnees[j] > LaDonneeCyto.Maxi[j] then
                    LaDonneeCyto.Maxi[j] := LaLigneDonnees[j]
                  else if (LaLigneDonnees[j] < LaDonneeCyto.Maxi[j]) and (LaLigneDonnees[j] > LaDonneeCyto.Maxi2[j]) then   //LaDonneeCyto.Maxi[j]*0.5
                    LaDonneeCyto.Maxi2[j] := LaLigneDonnees[j];
                end;
            end;
          Result := True;
        end;
    end;
end;

function TFCSFile.LitTextData : Boolean;
Var
  Code :Char;
  index,i,j : integer;
  np,np2,nd,nbpc : integer;
  ListeValeurs : TStringList;
  Keyword : string;
  Separator : Char;
  Subkey : string;
  TextData : string;
  Value :  AnsiString;//STRING;
  Erreur : string;
  ChParE,chaineVal : string;
  ListeNomsParamsComp : TStringList;
begin
  Result := False;
  // Process Text Segment
  TextData := GetTextData;

  // First character shows separation character for parsing
  Separator := TextData[1];
  Delete(TextData, 1, 1);

  SetLength(FName ,0);
  SetLength(FNameString, 0);
  SetLength(FRange,0);
  SetLength(FMask,0);
  SetLength(FBits ,0);
  SetLength(ParamCompN,0);
  SetLength(SpillMatrix,0);
  SetLength(InvSpillMatrix,0);
  SetLength(CMatrix,0);

  While Length(TextData) > 0 do
    begin
      index := POS(Separator, TextData);
      Keyword := COPY(TextData, 1, index-1);
      Delete(TextData, 1, index);

      index := POS(Separator,TextData);
      if index = 0 then
        begin
          Value := TextData;
          TextData := ''
        end
      else
        begin
          Value := COPY(TextData,1,index-1);
          Delete(TextData,1,index);
        end;
      //Lecture de la matrice de correction
      if (KeyWord = 'SPILL') or (KeyWord = '$SPILL') or (KeyWord = '$SPILLOVER')
           or (KeyWord = 'SPILLOVER')then  // or (KeyWord = '$COMP') or (KeyWord = 'COMP')   voir doc FCS3.0 dans ce cas
        begin
          np := pos(',',value);
          if np >0 then
            begin
              nbpc := 0;
              try
                nbpc := StrToInt(copy(value,1,np-1));
                if nbpc > 0 then
                  begin
                    Delete(value,1,np);
                    ListeValeurs := TStringList.Create;
                    try
                      ListeValeurs.StrictDelimiter := True;
                      ListeValeurs.Delimiter:=',';
                      ListeValeurs.DelimitedText:=AnsiToUTF8(Value);
                      SetLength(ParamCompN,nbpc);
                      SetLength(SpillMatrix,nbpc);
                      SetLength(InvSpillMatrix,nbpc);
                      {
                      FormMatrice.StringGrid1.RowCount:=nbpc+1;
                      FormMatrice.StringGrid1.ColCount:=nbpc+1;
                      FormMatrice.StringGrid2.RowCount:= nbpc+1;
                      FormMatrice.StringGrid2.ColCount:=nbpc+1;
                      }
                      np := ListeValeurs.Count;
                      if np = nbpc*(nbpc+1) then
                        begin
                          for i := 0 to nbpc-1 do //lignes
                            begin
                              ParamCompN[i]:= ListeValeurs[i];
                              //FormMatrice.StringGrid1.Cells[i+1,0]:=ParamCompN[i];
                              //FormMatrice.StringGrid1.Cells[0,i+1]:=ParamCompN[i];
                              SetLength(SpillMatrix[i],nbpc);
                              SetLength(InvSpillMatrix[i],nbpc);
                              for j := 0 to nbpc-1 do // colonnes
                                begin
                                  ChaineVal := ListeValeurs[nbpc*(i+1)+j];
                                  np := pos(NotSymbDec,ChaineVal);
                                  if np > 0 then
                                    ChaineVal[np] := SymbDec;
                                  try
                                    SpillMatrix[i,j]:=StrToFloat(ChaineVal); //j colonne, i : ligne
                                    if i=j then
                                      InvSpillMatrix[i,j]:=1
                                    else
                                      InvSpillMatrix[i,j]:=0;
                                    //FormMatrice.StringGrid1.Cells[j+1,i+1] := FloatToStr(SpillMatrix[i,j]);
                                    //FormMatrice.StringGrid2.Cells[j+1,i+1] := FloatToStr(InvSpillMatrix[i,j]);
                                  except
                                    SpillMatrix[i,j]:=0;
                                  end;
                                end;
                            end;
                        end;
                      //FormMatrice.ShowModal;
                      Inverse_LU;
                      //InvertSpillMatrix;
                     { for i := 0 to nbpc-1 do
                        for j := 0 to nbpc-1 do
                          FormMatrice.StringGrid2.Cells[j+1,i+1] := FloatToStr(InvSpillMatrix[i,j]);
                      FormMatrice.ShowModal;}
                    finally
                      ListeValeurs.Free;
                    end;
                  end;
              except
              end;
            end;
        end;
      if KeyWord = 'NOMPARAMSERIE' then
        FNomParamSerie := AnsiToUTF8(Value);
      if KeyWord = 'UNITPARAMSERIE' then
        FUniteParamSerie := AnsiToUTF8(Value);
      if KeyWord = 'VALPARAMSERIE' then
        try
          FValeurParamSerie := StrToInt(Value);
        except
          FValeurParamSerie := 0;
        end;
      if Keyword = '$PROJ' then
        FProj := AnsiToUTF8(Value);

      if Keyword = '$NEXTDATA' then
        begin
          try
            nd := StrToInt(Trim(Value));
            if nd >0 then
              FNextData := nd
            else
              FNextData := 0;
          except
            FNextData := 0;
          end;
        end;

      if Keyword = '$DATATYPE' then
        begin
          FDATATYPE := Value;
          if FDATATYPE = 'I' then
            FDataElementSize := 2
          else if FDATATYPE = 'F' then    // New 15 Feb 2006
            FDataElementSize := 4
        end;

      if Keyword = '$MODE' then
        FMODE := Value;

      if Keyword = '$FIL' then
        FFile := AnsiToUTF8(value);

      if Keyword = '$CELLS' then
        CELLS := AnsiToUTF8(Value);

      if Keyword = '$CYT' then
        Cytometer := Value;

      if Keyword = '$DATE' then
        LaDate := Value;

      if Keyword = '$TOT' then
        try
          EventCount := StrToInt(Trim(Value));
        except
          EventCount := 0;
          Erreur := 'Le nombre de cellules est illisible';
        end;

      if Keyword = '$BYTEORD' then
        begin
          BYTEORD := Value;
          if (Trim(Value) = '4,3,2,1') or (Trim(Value) = '2,1,4,3') then
            FBigEndian := True
          else
            begin
              BYTEORD :='1,2,3,4';
              FBigEndian := False;
            end;
        end;

      if Keyword = '$PAR' then
        begin
          ParameterCount := StrToInt(Trim(Value));
          SetLength(FName  ,ParameterCount);
          SetLength(FNameString,ParameterCount);
          SetLength(FRange ,ParameterCount);
          SetLength(FMask ,ParameterCount);
          SetLength(FBits  ,ParameterCount);
          SetLength(FParamE1 ,ParameterCount);
          SetLength(FParamE2 ,ParameterCount);
        end;

      if COPY(Keyword,1,2) = '$P' then
        begin
          Subkey := Keyword;
          Delete(Subkey,1,2);                  // get rid of $P
          Code := Subkey[ LENGTH(Subkey) ];
          Delete(Subkey, LENGTH(Subkey), 1);   // get rid of last letter
          index := StrToIntDef(Subkey,-1);     // index should remain

          if ((Code = 'N') or (Code = 'R') or (Code = 'B') or (Code = 'S') or (Code = 'E')) and (index > 0) then
            begin
              if index > Parametercount then //dans le cas où $PAR se trouve après
                begin
                  Parametercount := index;
                  SetLength(FName  ,ParameterCount);
                  SetLength(FNameString,ParameterCount);
                  SetLength(FRange ,ParameterCount);
                  SetLength(FMask ,ParameterCount);
                  SetLength(FBits  ,ParameterCount);
                  SetLength(FParamE1 ,ParameterCount);
                  SetLength(FParamE2 ,ParameterCount);
                end;
              Case code of
                'N': FName[index-1] := AnsiToUTF8(value);
                'R': FRange[index-1] := StrToInt(Trim(value));
                'S': FNameString[index-1] := AnsiToUTF8(value);
                'B': FBits[index-1]  := StrToInt(Trim(value));
                'E': begin
                       Value := Trim(Value);
                       np := pos(',',Value);
                       ChParE := Copy(Value,1,np-1);
                       Delete(Value,1,np);
                       try
                         np := pos(NotSymbdec,ChParE);
                         if np > 0 then
                           ChParE[np] := Symbdec;
                         FParamE1[index-1] := Round(StrToFloat(ChParE));

                         np := pos(NotSymbdec,Value);
                         if np > 0 then
                           Value[np] := Symbdec;
                         FParamE2[index-1] := StrToFloat(Value);
                       except
                         FParamE1[index-1] := 0;
                         FParamE2[index-1] := 0;
                       end;
                     end;
              end;
            end;
        end;
    end;
  if Length(ParamCompN) > 0 then     //construction de la matrice de correction
    begin
      ListeNomsParamsComp := TStringList.Create;
      try
        for i := 0 to Length(ParamCompN)-1 do
          ListeNomsParamsComp.Add(ParamCompN[i]);
        SetLength(CMatrix,ParameterCount);
        for i := 0 to ParameterCount-1 do
          begin
            SetLength(CMatrix[i],ParameterCount);
            np := ListeNomsParamsComp.IndexOf(FName[i]);
            if np = -1 then
              begin
                for j := 0 to ParameterCount-1 do
                  if j=i then
                    CMatrix[i,j]:=1
                  else
                    CMatrix[i,j]:=0;
              end
            else
              begin
                for j := 0 to ParameterCount-1 do
                  begin
                    np2 := ListeNomsParamsComp.IndexOf(FName[j]);
                    if np2 = -1 then
                      CMatrix[i,j]:=0
                    else
                      CMatrix[i,j]:=InvspillMatrix[np,np2];
                  end;
              end;
          end;
      finally
        ListeNomsParamsComp.Free;
      end;
    end;
  if FDATATYPE = '' then
    FDATATYPE := 'I';
  if BYTEORD = '' then
    begin
      BYTEORD := OrdreBitsDefaut;
      FBigEndian := InverserBits;
    end;
  if Erreur ='' then
    begin
      if (FDATATYPE = 'I') and (ParameterCount > 0) then
        FDataElementSize := FBits[0] div 8;
      For i := 0 to ParameterCount - 1 do
        if FRange[i] > 0 then
          FMask[i] := FRange[i]-1
        else
          FMask[i] := $FFFF;
      end;
      if (FHeaderDataLast > FHeaderDataFirst + (EventCount * FDataElementSize * ParameterCount)-1) then
        FHeaderDataLast := FHeaderDataFirst + (EventCount * FDataElementSize * ParameterCount)-1;
      if (FHeaderDataLast <> FHeaderDataFirst + (EventCount * FDataElementSize * ParameterCount)-1) then
        Erreur := 'Problème de déclaration du segment de données'+Chr(10)+chr(13)
                    + 'Le logiciel ne lit pas ce type de fichier';
      if (FDATATYPE <> 'I') and (FDATATYPE <> 'F') then
        Erreur := 'Le type de données n''est pas lisible par ce logiciel';
      if FMODE <> 'L' then
        Erreur :='Le type de données n''est pas lisible par ce logiciel';

  if Erreur <>'' then
    begin
      MessageDlg(Erreur,mtError,[mbOK],0);
      Version := 'Erreur';
    end
  else
    Result := True;
end;
Constructor TFCSFile.Create(Const filename:  string);
Const
  HeaderSize = 58;  // bytes minimum size (assumes no user-defined segment)
Var
  Code : Char;
  index ,i : integer;
  np : integer;
  Keyword : string;
  Separator : Char;
  Subkey :  STRING;
  TextData       :  STRING;
  Value          :  AnsiString;//STRING;
  Erreur : String;
begin
  OffsetGen := 0;
  //FBigEndian := False;    // Suppose endianess comme pour les PC (i.e., little endian)
  //BYTEORD := '1,2,3,4';
  FBigEndian := InverserBits;    // Suppose non endianess comme pour les mac
  BYTEORD := OrdreBitsDefaut;
  ParameterCount := 0;
  FDataElementSize := 2;  // Assume 2 bytes/element
  FNomParamSerie :='';
  FUniteParamSerie :='';
  FValeurParamSerie := 0;

  FFileName := filename;
  FFileStream := TFileStream.Create(FFilename, fmOpenRead);
  FNextData :=0;
{  if LitJeuDonnees then
    While (FNextData>0) and (FFileStream.Position < FFileStream.Size) do
      LitJeuDonnees;
      }
end; {Create}


Destructor TFCSFile.Destroy;
begin
  FFileStream.Free
end; {Destroy}


function TFCSFile.GetTextData : string;
Var
  TextSize : integer;
Begin
  TextSize := FHeaderTextLast - FHeaderTextFirst;
  SetLength(RESULT, TextSize);
  FFileStream.Seek(OffsetGen+FHeaderTextFirst, soFromBeginning);
  FFileStream.Read(RESULT[1], TextSize);
end;

FUNCTION TFCSFile.GetLigneDonnees(Const Row : integer;Const ApplyMasking, FixEndianess:  Boolean): TLigneDonnees;
  Var
    ElementSize : integer;
    i,j : integer;
    RowStart : integer;
    Value : Word;
    Value4 : LongWord;
    ValeurB : Byte;
    ShortFloat : Single  ABSOLUTE Value4;  // Memory overlay trick
    CopieLigne : TLigneDonnees;
begin
  ElementSize := FDataElementSize;  // 2 for 'I' or 4 for 'F'
  SetLength(Result,ParameterCount);
  // "Normal" data rows
  RowStart := FHeaderDataFirst + (Row - 1) * ElementSize * ParameterCount;

  // Caution:  Indexing is a bit confusing here
  FOR i := 1 TO ParameterCount DO
    BEGIN
      FFileStream.Seek(OffsetGen+RowStart+ElementSize*(i-1), soFromBeginning);
      IF   FDATATYPE = 'I' THEN
        BEGIN
          if ElementSize = 1 then
            begin
              FFileStream.Read(ValeurB,1);
              RESULT[i-1] := ValeurB;
            end
          else if ElementSize = 2 then
            begin
              FFileStream.Read(Value,2);
              if FBigEndian then
                Value:=Swap(Value);
              Value := Value AND FMask[i-1];
              RESULT[i-1] := Value;
            end
          else if ElementSize = 4 then
            begin
              FFileStream.Read(Value4,4);
              if FBigEndian then
                Value4:=Swap(Value4);
              Value4 := Value4 AND FMask[i-1];
              RESULT[i-1] := Value4;
            end;
          // Mask off extra bits that should be 0 but are not for some reason.
          // For example if Range is 4096, the mask = 4095 = $0FFF;
          // if Range is 1024, the mask = 1023 = $03FF
        END;

      IF FDATATYPE = 'F' THEN
        BEGIN
          FFileStream.Read(Value4, ElementSize);

          IF FBigEndian THEN
            Value4 := Swap4(Value4);

          RESULT[i-1] := ShortFloat;
          if Result[i-1] < 0 then
            Result[i-1] := 0;
        END;
  end;
  if Length(CMatrix)= ParameterCount then  // On tient compte de la matrice de correction
    begin
      SetLength(CopieLigne,ParameterCount);
      for i := 0 to ParameterCount-1 do
        CopieLigne[i] := Result[i];
      for i := 0 to ParameterCount-1 do
        begin
          Result[i] := CopieLigne[0]* CMatrix[0,i];
          for j := 1 to ParameterCount-1 do
            Result[i] := Result[i]+CopieLigne[j]* CMatrix[j,i];
        end;

    end;
end;
{
procedure TFCSFile.InvertSpillMatrix;
var
  arow: integer;
begin
  for arow := 0 to Length(SpillMatrix)-1 do
    populatediagonal(arow);
  for arow := 0 to Length(SpillMatrix)-1 do
    solveleft(arow);
  for arow := 0 to Length(SpillMatrix)-1 do
    solvecell(arow,arow,1);
  for arow := Length(SpillMatrix)-1 downto 0 do
    solveright(arow);
  for arow := 0 to Length(SpillMatrix)-1 do
    solvecell(arow,arow,1);
end;
function isdiff(a,b: real): boolean;
begin
  isdiff := (abs(a-b)>epsilon);
end;

procedure TFCSFile.addrows(r1,r2: integer; destrow: integer);
var
  acol: integer;
begin
  for acol := 0 to Length(SpillMatrix)-1 do   //décaler de -1     0 à Length(SpillMatrix)-1
    begin
      SpillMatrix[destrow,acol] := SpillMatrix[r1,acol]+SpillMatrix[r2,acol];
      InvSpillMatrix[destrow,acol] := InvSpillMatrix[r1,acol]+InvSpillMatrix[r2,acol];
    end;
end;

procedure TFCSFile.multrow(adouble: real; destrow: integer);
var
  acol: integer;
begin
  for acol := 0 to Length(SpillMatrix)-1  do
    begin
      SpillMatrix[destrow,acol] := SpillMatrix[destrow,acol]*adouble;
      InvSpillMatrix[destrow,acol] := InvSpillMatrix[destrow,acol]*adouble;
    end;
end;

procedure TFCSFile.solvecell(arow,acol: integer; desired: real);
var
  otherrow: integer;
  canuse: boolean;
  othercol: integer;
  Monmessage : string;
begin
  if isdiff(SpillMatrix[arow,acol],desired) then
    begin
      if isdiff(SpillMatrix[arow,acol],0) and (desired<>0) then
        multrow(1/SpillMatrix[arow,acol],arow);
      if isdiff(SpillMatrix[arow,acol],desired) then
        begin
          for otherrow := 0 to Length(SpillMatrix)-1 do
            if (otherrow<>arow) and isdiff(SpillMatrix[otherrow,acol],0) and
             isdiff(SpillMatrix[arow,acol],desired) then
              begin
                canuse := true;
                for othercol := 0 to acol-1 do
                  if isdiff(SpillMatrix[otherrow,othercol],0) then
                    canuse := false;
                if canuse then
                  begin
                    multrow(-SpillMatrix[otherrow,acol]/SpillMatrix[arow,acol],arow);
                    addrows(otherrow,arow,arow);
                  end;
              end;
        end;

      if isdiff(SpillMatrix[arow,acol],desired) then
        begin
          MonMessage :='Erreur dans l''inversion de la matrice ligne : '+IntToStr(arow)+' ; colonne : '
          +IntToStr(acol)+' valeur : ' + FloatToStr(SpillMatrix[arow,acol])+ ' recherché : '+ FloatToStr(desired);
          Messagedlg(MonMessage,mtWarning,[mbOK],0);
        end;
    end;
end;

procedure TFCSFile.populatediagonal(arow: integer);
var
  otherrow: integer;
begin
  if not isdiff(SpillMatrix[arow,arow],0) then
    begin
      for otherrow := 0 to Length(SpillMatrix)-1 do
        if isdiff(SpillMatrix[otherrow,arow],0) then
          if not isdiff(SpillMatrix[arow,arow],0) then
            addrows(otherrow,arow,arow);
    end;
end;

procedure TFCSFile.solveleft(arow: integer);
var
  acol: integer;
begin
  for acol := 0 to arow-1 do
    solvecell(arow,acol,0);
end;

procedure TFCSFile.solveright(arow: integer);
var
  acol: integer;
begin
  for acol := Length(SpillMatrix)-1 downto arow+1 do
    solvecell(arow,acol,0);
end;
}
procedure TFCSFile.Inverse_LU;
Var
  A : pVECT;      { matrix 0:n x 0:n stored in a vector (see NOTA2) }
  A1: pVect;      { copy of matrix A }
  Y : pVECT;      { matrix 0:n x 0:n stored in a vector }
  temp : pVECT;   { vector 0:n }
  INDX : pIVECT;  { integer vector 0:n }

  {NOTA1: index zero is not used here.
   NOTA2: The element i,j of matrix A(n,n) is the element i*(n+1)+j
          of the vector A, n being the 2nd dimension of A(n,n) }
  d, i, j, n, rc : integer;
begin
  New(A); New(A1); New(Y); New(temp); New(INDX);
  n := Length(SpillMatrix);
  for i:=1 to n do
  begin
    //ReadVec1(F1,n,n,temp);  {read a line}
    for j:=1 to n do
      begin
        A^[i*(n+1)+j] := SpillMatrix[i-1,j-1];//temp^[j];
        A1^[i*(n+1)+j] := A^[i*(n+1)+j];//temp^[j];
        Y^[i*(n+1)+j] := 0.0
      end;
    Y^[i*(n+1)+i] := 1.0;
    //WriteVec1(F2,n,n,temp)   {write a line}
  end;
 // close(F1);

{call LU decomposition routine (only once) }
  LUDCMP(A,n,INDX,D,rc);

{call solver if previous return code is ok
 to obtain inverse of A one column at a time}
  if rc=0 then
    for j:=1 to n do
      begin
        for i:=1 to n do
          temp^[i]:=Y^[i*(n+1)+j];
        LUBKSB(A,n,INDX,temp);
        for i:=1 to n do
          Y^[i*(n+1)+j]:=temp^[i];
      end;
  {inverse of matrix A is now in matrix Y,
   matrix A is destroyed. }

  if rc=1 then
    MessageDlg('La matrice est singulière',mtWarning,[mbOK],0)
  else
    begin
      for i:=1 to n do
        begin
          for j:=1 to n do
            InvSpillMatrix[i-1,j-1]:=Y^[i*(n+1)+j];
        end;
    end;
  //verify A1 x Y = I (result put in A)
//  MatMult(A1,Y,A,n);
  //A should now contain identity matrix
end;


END.
